﻿-- =============================================
-- Author:		Gerald Britton
-- Create date:     July 2017
-- Description:	Build table comparison script

-- Sample test cases:
--  EXEC #foo 'stg.USInvestmentRaw_PROD_170303', 'stg.USInvestmentRaw_EDPP_170303', 'UniqueID' -- select * from #source
--  EXEC #foo 'Vision.ref.PortfolioFeed', 'Vision_LFM.ref.PortfolioFeed', 'PortfolioFeedID', 1

-- EXEC GetComparisonScript N'Original', N'Revised', N'CustId', 1, 1

-- =============================================
ALTER PROCEDURE dbo.GetComparisonScript 
    @Source sysname,
    @Target sysname,
    @KeyColumns nvarchar(256),
    @is_identity bit = 0,
    @Debug bit = 0

AS BEGIN
    SET NOCOUNT ON;
    
    /************************
      Verify tables exists
    ************************/

    IF OBJECT_ID(@Source, N'U') IS NULL BEGIN
        RAISERROR('Source table: %s not found.', 11, 1, @Source);
        RETURN 1;
    END;
    IF OBJECT_ID(@Target, N'U') IS NULL BEGIN
        RAISERROR('Target table: %s not found.', 11, 2, @Target);
        RETURN 2;
    END;

    -- Variables
    DECLARE @stmt NVARCHAR(MAX), @cols NVARCHAR(MAX);
    DECLARE @CRLF VARCHAR(2) = char(10);

    -- Build temp tables from sys.columns for each table

    -- First we need a template for dynamic SQL, since the server and database may be supplied

    DECLARE @SysColumnsTemplate NVARCHAR(MAX) = N'
        INSERT INTO #Table
        SELECT * 
        FROM sys.columns c
        WHERE c.object_id = OBJECT_ID(@Table)
          AND NOT(@is_identity = 0 AND c.is_identity = 1)
        ';

    /*******************************************
        Get sys.columns for the source table
    *******************************************/

    -- DROP Temp table if it exists and make a new one
    IF object_id(N'tempdb..#Source', N'U') IS NOT NULL   
        DROP TABLE #Source;
    IF @Debug = 1
	   print 'SELECT TOP (0) * INTO #Source FROM sys.columns;'
    SELECT TOP (0) * INTO #Source FROM sys.columns;

    -- Now, let's get the full name of sys.columns

    DECLARE @syscolumns sysname; 
    SELECT @syscolumns = ISNULL(instance + N'.', N'') + ISNULL(db + N'.', N'') + N'sys.columns'
    FROM (VALUES(parsename(@Source,3), parsename(@Source,4))) v(db, instance);

    -- Complete the statement and execute it
    SET @stmt = replace(@SysColumnsTemplate, 'sys.columns', @syscolumns);
    SET @stmt = replace(@stmt, '#Table', '#Source');
    IF @Debug = 1 PRINT @stmt;
    EXEC sp_executesql @stmt, N'@Table sysname, @is_identity bit', @Source, @is_identity;

    -- We don't want to compare the object_id or default_object_id columns, so we'll drop them
 
    ALTER TABLE #Source DROP COLUMN object_id, default_object_id;

    /*******************************************
        Get sys.columns for the target table
    *******************************************/

    -- DROP Temp table if it exists and make a new one
    IF object_id(N'tempdb..#Target', N'U') IS NOT NULL   
        DROP TABLE #Target;
    IF @Debug = 1
	   print 'SELECT TOP (0) * INTO #Target FROM sys.columns;'
    SELECT TOP (0) * INTO #Target FROM sys.columns;

    -- Now, let's get the full name of sys.columns

    --DECLARE @syscolumns sysname; 
    SELECT @syscolumns = ISNULL(instance + N'.', N'') + ISNULL(db + N'.', N'') + N'sys.columns'
    FROM (VALUES(parsename(@Target,3), parsename(@Target,4))) v(db, instance);

    -- Complete the statement and execute it
    SET @stmt = replace(@SysColumnsTemplate, 'sys.columns', @syscolumns);
    SET @stmt = replace(@stmt, '#Table', '#Target');
    IF @Debug = 1 PRINT @stmt;
    EXEC sp_executesql @stmt, N'@Table sysname, @is_identity bit', @Target, @is_identity;
    ALTER TABLE #Target DROP COLUMN object_id, default_object_id

    -- Make sure the table definitions match
    IF EXISTS (SELECT * FROM #Source EXCEPT SELECT * FROM #Target) OR
       EXISTS (SELECT * FROM #Target EXCEPT SELECT * FROM #Source)
    BEGIN
        RAISERROR('Tables do not match.', 11, 3);
        RETURN 3;
    END;

    /***************************************************
        -- Create join predicate from key columns --
    ***************************************************/

    DECLARE @Predicate NVARCHAR(4000);
    DECLARE @Tokens TABLE (Token VARCHAR(255));
    INSERT INTO @Tokens(token)
        SELECT token FROM Split(@KeyColumns, ',')

    SET @predicate = (
        SELECT STUFF((
             SELECT ' and src.' + token + ' = tgt.' + token
             FROM @Tokens
             FOR XML PATH('')), 
            1, 5, '')
    )

    /***************************************************
        -- Create key columns for select --
    ***************************************************/

    DECLARE @SelectKeys NVARCHAR(4000) = 
    (
        SELECT STUFF((
             SELECT ', src.' + token
             FROM @Tokens
             FOR XML PATH('')), 
            1, 2, '')
    )


    -- get difference columns
    SET @stmt = N'
        IF  OBJECT_ID(N''tempdb..#diffcols'', N''U'') IS NOT NULL
            DROP TABLE #diffcols;

        SELECT COUNT(*) AS [Source Row Count] FROM {Source} src          
        SELECT COUNT(*) AS [Target Row Count] FROM {Target} tgt 

        SELECT 
            {KeyColumns},
            CONCAT({cols}) + '', ''
        AS cols
        INTO #diffcols
        FROM {Source} src
        JOIN {Target} tgt
          ON {Predicate}
        WHERE EXISTS (SELECT src.* EXCEPT SELECT tgt.*)
        ;
     ';

     SET @stmt = replace(@stmt, replicate(' ',8), '');
     -- DECLARE @FullTable1 NVARCHAR(132) = 'stg.usdemanddemandraw'; 
     DECLARE @maxnamelen INT = (SELECT MAX(LEN(c.name)) FROM #Source c);

    /* --
        DECLARE @FullTable1 NVARCHAR(132) = 'stg.usdemanddemandraw'; 
        DECLARE @cols nvarchar(max), @maxnamelen int = 20
    -- */
    SET @cols = (
     SELECT @crlf + REPLICATE(N' ', 4)
        + N'IIF(EXISTS(SELECT src.' 
        + c.name + REPLICATE(' ', @maxnamelen - LEN(c.name) + 1)
        + N' EXCEPT SELECT tgt.' 
        + c.name + REPLICATE(' ', @maxnamelen - LEN(c.name) + 1)
        + N'), RTRIM('', ' 
        + c.name + REPLICATE(' ', @maxnamelen - LEN(c.name) + 1)
        + N'''), ''''),' 
     FROM #Source c
     FOR XML PATH('')
    );

    --SELECT @cols

    SET @stmt = REPLACE(@stmt, N'{Source}'      , @Source);
    SET @stmt = REPLACE(@stmt, N'{Target}'      , @Target);
    SET @stmt = REPLACE(@stmt, N'{cols}'        , LEFT(@cols, LEN(@cols)- 1));
    SET @stmt = REPLACE(@stmt, N'{KeyColumns}'  , @SelectKeys);
    SET @stmt = REPLACE(@stmt, N'{Predicate}'   , @Predicate);

     DECLARE @start int = 1;
     WHILE @start <= len(@stmt) BEGIN
        PRINT substring(@stmt, @start, 4000);
        SET   @start += 4000;
     END;

	PRINT 'SELECT COUNT(*) AS [Difference Row Count] from #diffcols;'
    /* -- 
    DECLARE @FullTable1 NVARCHAR(132) = 'stg.usdemanddemandraw'; DECLARE @cols nvarchar(max) 
    DECLARE @maxnamelen INT = (SELECT MAX(LEN(c.name)) FROM sys.columns c WHERE c.object_id = OBJECT_ID(@FullTable1))
    --*/

    SET @stmt = N'
        WITH src AS (
            SELECT 
                {cols}
            FROM #diffcols d
        )
        SELECT ca.col AS ColumnName, ca.diff AS [Difference Count]
        FROM src
        CROSS APPLY ( VALUES
            {values}
        ) ca(col, diff)
        WHERE diff > 0
        ORDER BY diff desc
        ;
     ';
    SET @stmt = replace(@stmt, replicate(' ',8), '');
    
    SET @cols = (
     SELECT @crlf + REPLICATE(N' ', 4)
        + 'SUM(IIF(d.cols LIKE ''%, ' 
        + c.name + ', %'''
        + REPLICATE(' ', @maxnamelen - LEN(c.name) + 1)
        + ' , 1, 0)) AS ' 
        + c.name
        + ','
     FROM #Source c
     FOR XML PATH('') 
    );

    SET @stmt = REPLACE(@stmt, N'{Source}', @Source);
    SET @stmt = REPLACE(@stmt, N'{Target}', @Target);
    SET @stmt = REPLACE(@stmt, N'{cols}'  , LEFT(@cols, LEN(@cols)- 1));

    /* -- 
    DECLARE @FullTable1 NVARCHAR(132) = 'stg.usdemanddemandraw'; DECLARE @cols nvarchar(max) 
    DECLARE @maxnamelen INT = (SELECT MAX(LEN(c.name)) FROM sys.columns c WHERE c.object_id = OBJECT_ID(@FullTable1))
    --*/

    SET @cols = (
     SELECT @crlf + REPLICATE(N' ', 4)
        + '('''
        + c.name + REPLICATE(' ', @maxnamelen - LEN(c.name) + 1)
        + ''',' 
        + c.name + REPLICATE(' ', @maxnamelen - LEN(c.name) + 1)
        + '),'
     FROM #Source c
     FOR XML PATH('')
    );

     SET @stmt = REPLACE(@stmt, N'{values}', LEFT(@cols, LEN(@cols)- 1));

     SET @start = 1;
     WHILE @start <= len(@stmt) BEGIN
        PRINT substring(@stmt, @start, 4000);
        SET   @start += 4000;
     END;
END
